"""Core constants."""

__all__ = (
    'Constants',
    )

from . import lib
from . import typ


class Constants:
    """Constant values shared across all of fqr."""

    PACAKGE = 'fqr'
    """Reference to package name."""

    CLASS_AS_DICT: typ.string[typ.snake_case] = 'class_as_dict'
    """Common string reference."""

    DELIM = '-'
    """Common delimiter used throughout package."""

    INDENT = int(lib.os.getenv('INDENT', 2))
    """
    Default, package-wide indentation.

    ---

    Used for `__repr__`, log messages, etc.

    """

    UNDEFINED = f'[[{PACAKGE.upper()}_DEFAULT_PLACEHOLDER]]'
    """Placeholder for undefined values that should not be `None`."""

    DELIM_REBASE = '_X_'

    __ANNOTATIONS__: typ.string[typ.snake_case] = '__annotations__'
    __DATACLASS_FIELDS__: typ.string[typ.snake_case] = '__dataclass_fields__'
    __DICT__: typ.string[typ.snake_case] = '__dict__'
    __HERITAGE__: typ.string[typ.snake_case] = '__heritage__'
    __SLOTS__: typ.string[typ.snake_case] = '__slots__'
    __MODULE__: typ.string[typ.snake_case] = '__module__'

    FIELDS: typ.string[typ.snake_case] = 'fields'
    ENUMERATIONS: typ.string[typ.snake_case] = 'enumerations'
    HASH_FIELDS: typ.string[typ.snake_case] = 'hash_fields'



"""Typing objects."""

__all__ = (
    'ArrayProto',
    'FieldPattern',
    'MappingProto',
    'MetaLike',
    'ObjectLike',
    'SupportsAnnotations',
    'SupportsParams',
    'VariadicArrayProto',
    'WrapperPattern',
    )

from .. import typ

from . import cns
from . import lib


class Constants(cns.Constants):
    """Constant values specific to this file."""


FieldPattern = lib.re.compile(
    r'(fqr(\.[a-zA-Z]{1,32}){0,32}\.)?Field'
    r'\[((\[)?[\.\|\,a-zA-Z0-9_ ]{1,64}(\])?){1,64}\]'
    )

WrapperPattern = lib.re.compile(
    r'([a-zA-Z]{1,64}\.?)?(Annotated|ClassVar|Final|InitVar)'
    r'\[((\[)?[\.\|\,a-zA-Z0-9_ ]{1,64}(\])?){1,64}\]'
    )


class ArrayProto(lib.t.Protocol, lib.t.Collection[typ.AnyTypeCo]):
    """Protocol for a generic, single-parameter array."""

    def __init__(
        self,
        iterable: lib.t.Iterable[typ.AnyTypeCo],
        /
        ) -> None: ...

    def __iter__(self) -> lib.t.Iterator[typ.AnyTypeCo]: ...


class VariadicArrayProto(
    ArrayProto[tuple[lib.Unpack[typ.ArgsType]]],
    lib.t.Protocol
    ):
    """Protocol for a generic, any-parameter array."""

    def __hash__(self) -> int: ...


class MappingProto(
    lib.t.Protocol,
    lib.t.Generic[typ.AnyTypeCo, typ.AnyOtherTypeCo]
    ):
    """Protocol for a generic, double-parameter mapping."""

    def __init__(self, *args: lib.t.Any, **kwargs: lib.t.Any) -> None: ...

    def __iter__(self) -> lib.t.Iterator[typ.AnyTypeCo]: ...

    def __getitem__(
        self,
        __name: str,
        __default: lib.t.Optional[typ.AnyType] = None
        ) -> typ.AnyTypeCo | typ.AnyType: ...

    def items(self) -> lib.t.ItemsView[typ.AnyTypeCo, typ.AnyOtherTypeCo]: ...

    def keys(self) -> lib.t.KeysView[typ.AnyTypeCo]: ...

    def values(self) -> lib.t.ValuesView[typ.AnyOtherTypeCo]: ...


class SupportsAnnotations(lib.t.Protocol):
    """
    Protocol for a typed object.

    ---

    Typed objects include `dataclass`, `TypedDict`, `pydantic.Model`, \
    and both `fqr.Field` and `fqr.Object` amongst others.

    """

    __annotations__: dict[str, lib.t.Any]
    __bases__: tuple[type, ...]
    __name__: str

    def __init__(self, *args: lib.t.Any, **kwargs: lib.t.Any) -> None: ...


class SupportsParams(lib.t.Protocol, lib.t.Generic[lib.Unpack[typ.ArgsType]]):
    """Protocol for a generic with any number of parameters."""

    if lib.sys.version_info >= (3, 9):
        def __class_getitem__(
            cls,
            item: tuple[lib.Unpack[typ.ArgsType]],
            /
            ) -> lib.types.GenericAlias: ...

    __args__: tuple[lib.Unpack[typ.ArgsType]]

    def __hash__(self) -> int: ...


class MetaLike(lib.t.Protocol):
    """Meta protocol."""

    __annotations__: typ.SnakeDict
    __dataclass_fields__: lib.t.ClassVar[typ.DataClassFields]


class ObjectLike(lib.t.Protocol):
    """Object protocol."""

    __annotations__: typ.SnakeDict
    __dataclass_fields__: lib.t.ClassVar[typ.DataClassFields]

    def __contains__(self, __key: lib.t.Any, /) -> bool: ...

    def __getitem__(self, __key: lib.t.Any, /) -> lib.t.Any: ...

    def __setitem__(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]: ...

    def __ior__(self, other: 'ObjectLike', /) -> None: ...

    def get(
        self,
        __key: typ.AnyString,
        __default: typ.AnyType = None
        ) -> lib.t.Any | typ.AnyType: ...

    def items(
        self
        ) -> lib.t.ItemsView[typ.string[typ.snake_case], lib.t.Any]: ...

    @classmethod
    def keys(cls) -> lib.t.KeysView[typ.string[typ.snake_case]]: ...

    def pop(
        self,
        __key: str,
        /,
        __default: typ.AnyType = Constants.UNDEFINED
        ) -> typ.AnyType | lib.t.Any | lib.Never: ...

    def setdefault(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]: ...

    def update(self, other: 'ObjectLike', /) -> None: ...

    def values(self) -> lib.t.ValuesView[lib.t.Any]: ...

    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[False] = False,
        include_null: bool = True
        ) -> typ.SnakeDict: ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[True],
        include_null: bool
        ) -> typ.CamelDict: ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: bool,
        include_null: bool
        ) -> 'typ.SnakeDict | typ.CamelDict': ...
    def to_dict(
        self,
        camel_case: bool = False,
        include_null: bool = True
        ) -> 'typ.SnakeDict | typ.CamelDict': ...



"""Typing types."""

from .. import typ

__all__ = (
    'AnyOrForwardRef',
    'Array',
    'ArrayType',
    'Immutable',
    'Mapping',
    'MappingType',
    'NumberType',
    'Typed',
    'VariadicArray',
    'VariadicArrayType',
    'StrOrForwardRef',
    'UnionGenericAlias',
    *typ.__all__
    )

from .. typ import *

from . import lib
from . import obj

AnyOrForwardRef = lib.t.ForwardRef | lib.t.Any
StrOrForwardRef = lib.t.ForwardRef | str
UnionGenericAlias = type(int | str)
Wrapper = obj.SupportsParams[lib.Unpack[ArgsType]]

Array: lib.t.TypeAlias = obj.ArrayProto[AnyType]
Immutable = (
    bool
    | bytes
    | complex
    | lib.decimal.Decimal
    | lib.enum.Enum
    | float
    | lib.fractions.Fraction
    | int
    | NoneType
    | range
    | str
    | lib.enum.EnumMeta
    | frozenset['Immutable']
    | tuple['Immutable', ...]
    | lib.types.MappingProxyType['Immutable', 'Immutable']
    )
Mapping: lib.t.TypeAlias = obj.MappingProto[AnyType, AnyOtherType]
NumberType = lib.t.TypeVar('NumberType', bound=lib.numbers.Number)
Typed = obj.SupportsAnnotations
VariadicArray: lib.t.TypeAlias = (
    obj.VariadicArrayProto[lib.Unpack[tuple[AnyType, ...]]]
    )

ArrayType = lib.t.TypeVar('ArrayType', bound=Array)
MappingType = lib.t.TypeVar('MappingType', bound=Mapping)
VariadicArrayType = lib.t.TypeVar('VariadicArrayType', bound=VariadicArray)



"""Module objs unit tests."""

import unittest

import fqr

from fqr . core import lib

from ... import mocking

from . import cns


class Constants(cns.Constants):
    """Constant values specific to unit tests in this file."""


class TestObject(unittest.TestCase):
    """Fixture for testing `Object`."""

    def setUp(self) -> None:
        self.mcs = fqr.objects.metas.Meta
        self.cls = mocking.Derivative
        self.object_ = self.cls()
        self.trip = mocking.TripDeriv(
            str_field='123',
            )
        self.anti = mocking.AntiTripDeriv(
            str_field='321'
            )
        return super().setUp()

    def test_01_dict_functionality(self):
        """Test `Object.__getitem__()`."""

        self.assertTrue(
            self.object_['int_field']
            == self.object_.int_field
            == self.cls.int_field.default
            )

    def test_02_dict_functionality(self):
        """Test `Object.__getitem__()` raises KeyError if no key."""

        self.assertRaises(
            KeyError,
            lambda: self.object_['field_that_does_not_exist']
            )

    def test_03_iter(self):
        """Test `Object.__iter__()`."""

        self.assertEqual(dict(self.object_.__iter__()), dict(self.object_))

    def test_04_len(self):
        """Test `Object.__len__()`."""

        self.assertEqual(len(self.object_), len(self.cls.fields))

    def test_05_contains(self):
        """Test `Object.__contains__()`."""

        self.assertIn(self.object_.fields[0], self.object_)

    def test_06_ne(self):
        """Test `Object.__ne__()`."""

        self.assertFalse(self.object_ != self.object_)

    def test_07_lshift(self):
        """Test `Object.__lshift__()` correctly interpolates."""

        object_ = self.trip << self.anti
        self.assertNotEqual(object_.str_field, self.anti.str_field)

    def test_08_lshift(self):
        """Test `Object.__lshift__()` correctly interpolates."""

        object_ = self.trip << self.anti
        self.assertNotEqual(object_.other_field, self.trip.other_field)

    def test_09_rshift(self):
        """Test `Object.__rshift__()` correctly overwrites."""

        default = mocking.TripDeriv()
        object_ = self.trip >> default
        self.assertNotEqual(object_.str_field, default.str_field)

    def test_10_rshift(self):
        """Test `Object.__rshift__()` correctly overwrites."""

        object_ = self.trip >> self.anti
        self.assertEqual(object_.str_field, self.anti.str_field)

    def test_11_dict_functionality(self):
        """Test `Object.get()` returns default if no key."""

        self.assertIsNone(self.object_.get('field_that_does_not_exist'))

    def test_12_to_dict(self):
        """Test `Object.to_dict()`."""

        self.assertDictEqual(
            {
                fqr.core.strings.utl.snake_case_to_camel_case(k): self.trip[k]
                for k, v in self.trip.items()
                if type(v) in lib.t.get_args(fqr.core.typ.Primitive)
                },
            {
                k: v
                for k, v
                in self.trip.to_dict(camel_case=True).items()
                if type(v) in lib.t.get_args(fqr.core.typ.Primitive)
                }
            )

    def test_13_to_dict(self):
        """Test `Object.to_dict()`."""

        self.assertDictEqual(
            {k: v for k, v in self.object_ if v is not None},
            self.object_.to_dict(include_null=False)
            )

    def test_14_repr(self):
        """Test `Object.__repr__()`."""

        self.assertEqual(
            repr(self.trip),
            lib.json.dumps(
                dict(self.trip),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                )
            )

    def tearDown(self) -> None:
        return super().tearDown()



"""Core typing."""

__all__ = (
    'camelCase',
    'datetime',
    'numeric',
    'snake_case',
    'string',
    'AnyDict',
    'AnyField',
    'AnyOtherType',
    'AnyOtherTypeCo',
    'AnyString',
    'AnyType',
    'AnyTypeCo',
    'ArgsType',
    'CamelDict',
    'Casing',
    'DataClassFields',
    'Enum',
    'FieldsTuple',
    'Literal',
    'NoneType',
    'Object',
    'ObjectType',
    'OptionalAnyDict',
    'Primitive',
    'PackageExceptionType',
    'Serial',
    'SnakeDict',
    'StringFormat',
    'StringType',
    )

from . import lib

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import objects  # noqa: F401
    from . import exc  # noqa: F401
    from . import typings  # noqa: F401

camelCase = lib.t.NewType('camelCase', str)
snake_case = lib.t.NewType('snake_case', str)
datetime = lib.t.NewType('datetime', str)
numeric = lib.t.NewType('numeric', str)

AnyDict = dict[str, lib.t.Any]
AnyField: lib.t.TypeAlias = 'objects.Field[AnyType]'
AnyString: lib.t.TypeAlias = lib.t.Union[str, 'string[StringType]']
CamelDict: lib.t.TypeAlias = 'dict[string[camelCase], lib.t.Any]'
Casing = (
    camelCase
    | snake_case
    )
DataClassFields: lib.t.TypeAlias = 'dict[string[snake_case], AnyField[lib.t.Any]]'  # noqa
Enum: lib.t.TypeAlias = lib.t.Union[
    'typings.typ.Array[typings.typ.Immutable]',
    lib.enum.EnumMeta
    ]
Field: lib.t.TypeAlias = 'objects.Field'
FieldsTuple: lib.t.TypeAlias = 'tuple[string[snake_case], ...]'
Literal = lib.t.Literal['*']
NoneType = lib.types.NoneType  # type: ignore[valid-type]
Object: lib.t.TypeAlias = 'objects.Object'
OptionalAnyDict = lib.t.Optional[dict[str, lib.t.Any]]
Primitive = bool | float | int | NoneType | str
Serial = Primitive | dict[Primitive, 'Serial'] | list['Serial']
SnakeDict: lib.t.TypeAlias = 'dict[string[snake_case], lib.t.Any]'
StringFormat = (
    camelCase
    | snake_case
    | datetime
    | numeric
    )

AnyType = lib.t.TypeVar('AnyType')
AnyOtherType = lib.t.TypeVar('AnyOtherType')
AnyTypeCo = lib.t.TypeVar('AnyTypeCo', covariant=True)
AnyOtherTypeCo = lib.t.TypeVar('AnyOtherTypeCo', covariant=True)
ArgsType = lib.TypeVarTuple('ArgsType')
ObjectType = lib.t.TypeVar('ObjectType', bound=Object)
StringType = lib.t.TypeVar('StringType', bound=StringFormat)

PackageExceptionType = lib.t.TypeVar(
    'PackageExceptionType',
    bound='exc.BasePackageException',
    covariant=True,
    )


class string(str, lib.t.Generic[StringType]):
    """Protocol for a generic `str`."""
